Title:  Solution to FMFP'12 exercise sheet 1
Author: Ben Bittdiddle
==============================================

Assignment 1
------------

The complete solution is given in the input file sheet1_johndo.hs.


Assignment 2
------------

In Haskell, one may define synomyms for commonly used types.
This is done using the keyword "type".

Example:

> type Complex = (Float,Float)

With this definition we can use the shorthand "Complex" instead of 
"(Float, Float)". Note that this does NOT define a new type---it 
only gives a new name for an existing type.


(a) 

> re :: Complex -> Float
> re (x,y) = x

> im :: Complex -> Float
> im (x,y) = y


(b)

> conj :: Complex -> Complex
> conj (x,y)= (x,-y)


(c) 

Complex addition: componentwise

> add :: Complex -> Complex -> Complex
> add (r1,i1) (r2,i2) = (r1+r2,i1+i2)

Complex multiplication: (a+bi) (c+di) = (ac - bd) + (ad + cb)i

> mult :: Complex -> Complex -> Complex
> mult (r1,i1) (r2,i2) = (r1*r2-i1*i2,r1*i2+r2*i1)

Absolute value of a complex number:

> abs1 :: Complex -> Float
> abs1 (r,i) = sqrt (r*r+i*i)

Alternative definition using the conjugate:

> abs2 :: Complex -> Float
> abs2 x = sqrt (re (mult x (conj x)))



Assignment 3
------------

> fibLouis :: Int -> Int
> fibLouis 0 = 0                                -- fibLouis.1
> fibLouis 1 = 1                                -- fibLouis.2
> fibLouis n = fibLouis (n-1) + fibLouis (n-2)  -- fibLouis.3
>
> fibEva :: Int -> Int
> fibEva n = fst (aux n)                        -- fibEva.1
>     where aux 0 = (0,1)                       -- aux.1
>           aux n = next (aux (n-1))            -- aux.2
>           next (a,b) = (b, a+b)               -- next.1


The evaluation order in Haskell will be discussed in more detail
later. However, we want point out that it can make a difference how
you evaluate terms. Depending on the evaluation strategy, you can
save an infinite amonunt of work.

Haskell uses the following strategy for evaluating terms.
No argument is evaluated twice, and:
* Evaluation is from the outside in.
  Example: In a situation like
    f e (f' e')
  the outer one, f e (f' e') is chosen for evaluation.
* Otherwise, evaluation is from the left to the right.
  Example: In the expression
    f e + f' e'
  The left-hand sub-expression, f e will be evaluated first. 


For our examples, we get:

(a) 

fibLouis 4 =
(fibLouis (4-1) + fibLouis (4-2)) =
(fibLouis 3 + fibLouis (4-2)) =
((fibLouis (3-1) + fibLouis (3-2)) + fibLouis (4-2)) =
((fibLouis 2 + fibLouis (3-2)) + fibLouis (4-2)) =
(((fibLouis (2-1) + fibLouis (2-2)) + fibLouis (3-2)) + fibLouis (4-2)) =
(((fibLouis 1 + fibLouis (2-2)) + fibLouis (3-2)) + fibLouis (4-2)) =
(((1 + fibLouis (2-2)) + fibLouis (3-2)) + fibLouis (4-2)) =
(((1 + fibLouis 0) + fibLouis (3-2)) + fibLouis (4-2)) =
(((1 + 0) + fibLouis (3-2)) + fibLouis (4-2)) =
((1 + fibLouis (3-2)) + fibLouis (4-2)) =
((1 + fibLouis 1) + fibLouis (4-2)) =
((1 + 1) + fibLouis (4-2)) =
(2 + fibLouis (4-2)) =
(2 + fibLouis 2) =
(2 + (fibLouis (2-1) + fibLouis (2-2))) =
(2 + (fibLouis 1 + fibLouis (2-2))) =
(2 + (1 + fibLouis (2-2))) =
(2 + (1 + fibLouis 0)) =
(2 + (1 + 0)) =
(2 + 1) =
3

(b)

fibEva 4 = 
fst (aux 4) =
fst (next (aux (4-1))) = 
fst (next (aux 3)) = 
fst (next (next (aux (3-1)))) =
fst (next (next (aux 2))) =
fst (next (next (next (aux (2-1))))) =
fst (next (next (next (aux 1)))) =
fst (next (next (next (next (aux (1-1)))))) =
fst (next (next (next (next (aux 0))))) =
fst (next (next (next (next (0,1))))) =
fst (next (next (next (1,0+1)))) =
fst (next (next ((0+1),1+(0+1)))) =
fst (next (1+(0+1),(0+1)+(1+(0+1)))) =
fst ((0+1)+(1+(0+1)),(1+(0+1))+(0+1)+(1+(0+1))) = 
(0+1)+(1+(0+1)) =
1 + (1 + 1) =
1 + 2 =
3


NOTE that the running times of fibEva and fibLouis are quite different.

Try

  Main> fibEva 30

and

  Main> fibLouis 30

The difference in running times can best be explained when you write the
evaluation of fibLouis as a tree:


         fibL 8
        |     \
     fibL 7   fibL 6
     |   \      |  \
fibL 6  fibL 5 ... ...
    |     \
  ...    ...


The entire subtree for fibLouis 6 is calculated twice,
the subtree for fibLouis 5 is calculated three times,
the subtree for fibLouis 4 is calculated 5 times, ...

This leads to a blow-up in running times compared to fibEva, where aux is
evaluated only once for every value.  In summary, fibLouis has a running time
that is exponential in n, i.e., O(2^n). The running time of fibEva is linear
in n, i.e., O(n).


Assignment 4, 5
---------------

The solutions to these two assignments are given in the file solution1.pdf.
