SOLUTION TO EXERCISE SHEET 4 FMFP'12
-------------------------------------

Author: Whizbang Weasle


Assignment 1:
-------------

> data Number = N Int [Int]

  elegant:

> toInt :: Number -> Int
> toInt (N base ds) = sum . zipWith (*) (iterate (base*) 1) $ ds

  more to the specification:

> toInt' :: Number -> Int
> toInt' (N base ds) = sum . zipWith (*) (map (base^) [0..]) $ ds

  Note that the operator ($) :: (a -> b) -> a -> b denotes application; i.e.  
  f $ a = f a. The central difference is that it has a very low precedence and
  associates to the right. Often it allows to save on parentheses thus making
  parsing of an expression easier.


Assignment 2:
------------

Recall the definitions from the assignment.

  rev :: [a] -> [a]
  rev []     = []                   -- rev.1
  rev (x:xs) = rev xs ++ [x]        -- rev.2

  (++) :: [a] -> [a] -> [a]
  (++) []     ys = ys               -- app.1
  (++) (x:xs) ys = x : (xs ++ ys)   -- app.2

Lemma: Forall ys::[a] xs::[a]. rev (xs ++ rev ys) = ys ++ rev xs

Proof: 
  Let ys::[a] be arbitrary and
  P(xs) := rev (xs ++ rev ys) = ys ++ rev xs. 
  We show Forall xs::[a]. P(xs) by induction.

  Base case: 
    Show P([]).
    Hence, we have to show rev ([] ++ rev ys) = ys ++ rev [].

      rev ([] ++ rev ys)  
    = rev (rev ys)        -- app.1  
    = ys                  -- Lemma 1
    = ys ++ []            -- Lemma 2
    = ys ++ rev []        -- rev.1   
                                   
  Step case:
    Let x::a and xs::[a] be arbitrary. Assume P(xs). Show P(x:xs).
    Hence, we have to show rev ((x:xs) ++ rev ys) = ys ++ rev (x:xs).

      rev ((x:xs) ++ rev ys)    
    = rev (x:(xs ++ rev ys))     -- app.2
    = rev (xs ++ rev ys) ++ [x]  -- rev.2
    = (ys ++ rev xs) ++ [x]      -- P(xs)
    = ys ++ ((rev xs) ++ [x])    -- Lemma 3
    = ys ++ rev (x:xs)           -- rev.2     

qed.


Assignment 3:
------------
(a)

> split sep s = aux s ""
>   where aux []     w = [w]
>         aux (c:cs) w | c == sep  = w : aux cs ""
>                      | otherwise = aux cs (w++[c])

> split_tests = [(split '#' "foo##goo",["foo","","goo"]),
>                (split '#' "#",["",""])]

(b)

> center :: [String] -> [String]
> center lines = map centerLine lines
>   where 
>   width = maximum (map length lines)
>   centerLine line = replicate left ' ' ++ line ++ replicate right ' '
>     where 
>     len   = length line
>     left  = (width - len) `div` 2
>     right = width - len - left

> center_tests = [(center ["fullness","of","exposition","is","necessary"],
>                  [" fullness ",
>                   "    of    ",
>                   "exposition",
>                   "    is    ",
>                   "necessary "])]


Assignment 4:
------------

(a) filterMap with foldr

> filterMap p f = foldr aux e
>   where aux = \x xs -> if p (f x) then f x : xs else xs
>         e   = []

(b) foldl with foldr

Solution tactic:

Note that foldr is isomorphic to primitive recursion. For example,

  sum []     = 0
  sum (x:xs) = x + sum xs

can also be written with an explicit combinator function for the recursion

  sum []     = 0
  sum (x:xs) = (\x folded -> x + folded) x (sum xs)

which makes the following foldr-based implementation obvious.

  sum = foldr (\x folded -> x + folded) 0

Hence, we just need to convert the implementation of foldl such that the primitive 
recursion becomes explicit. Let's start with the standard implementation of foldl.

> myFoldl' :: (b -> a -> b) -> b -> [a] -> b
> myFoldl' _ z []     = z
> myFoldl' f z (x:xs) = myFoldl' f (z `f` x) xs

The following first rewrite makes the primitive recursion almost obvious.

> myFoldl'' :: (b -> a -> b) -> b -> [a] -> b
> myFoldl'' f z0 xs0 = 
>     recurse xs0 z0
>   where
>     recurse []     z = z
>     recurse (x:xs) z = recurse xs (z `f` x)

Now, we just move the z-argument of recurse to the right-hand side and the 
primitive recursion becomes obvious.

> myFoldl''' :: (b -> a -> b) -> b -> [a] -> b
> myFoldl''' f z0 xs0 = 
>     recurse xs0 z0
>   where
>     recurse []     = \z -> z
>     recurse (x:xs) = \z -> (recurse xs) (z `f` x)

Finally, we replace primitive recursion by a call to foldr and we're done.

> myFoldl :: (b -> a -> b) -> b -> [a] -> b
> myFoldl f z0 xs = foldr (\x folded z -> folded (z `f` x)) id xs z0


Let's prove that this definition is the right one.
--------------------------------------------------

Claim: Forall xs z. foldl f z xs = foldr (\x r l-> r (f l x)) id xs z

Proof: 
  Let P (xs) := forall z. foldl f z xs = foldr g id xs z
  and g x r l := r (f l x)

  We show 'Forall xs::[a]. P(xs)' by induction.

  Base Case: Show that P [] holds.
    P [] <=> Forall z. foldl f z [] = foldr g id [] z  [unfold P]
         <=> foldl f z [] = foldr g id [] z            [all intro]
         <=> z = id z                                  [def: foldl, foldr]
	 <=> z = z                                     [def: id]
	 <=> True                                      [refl of = ]

  Step Case: Asumme P (xs) holds; i.e.
               Forall z. foldl f z xs = foldr g id xs z
             holds. Show that this implies P (x:xs)
  
    P (x:xs) <=> Forall z. foldl f z (x:xs) = foldr g id (x:xs) z   [unfold P]
             <=> foldl f z (x:xs) = foldr g id (x:xs) z             [all intro]
             <=> foldl f (f z x) xs = g x (foldr g id xs) z         [def: foldl, foldr]
             <=> foldl f (f z x) xs = foldr g id xs (f z x)         [def: g]
             <=> True                                               [P(xs) with z := (f z x)]

qed.


