-- Title:  Solution to FMFP'12 exercise sheet 7
-- Author: Håkon Jørgensen
-- ============================================


-- Assignment 1
---------------

> data Tree a = Leaf a | Node (Tree a) (Tree a)

a)

> instance Monad Tree where
>     return         = Leaf                         -- return.1
> 
>     Leaf x >>= f   = f x                          -- (>>=).1
>     Node l r >>= f = Node (l >>= f) (r >>= f)     -- (>>=).2


b) 

Note that constructing proofs of similar difficulty as the ones below
might be a question in the exam. If you haven't solved question (a) then
use the above solution and try to solve question (b) based on it.

Once you've solved part (b), compare to the following proofs. Notice the
explicit statement of claims and definitions and the explicit elimination
of universal quantification using constructions like "Let x::a be
arbitrary". This makes proof easier to read and harder to get wrong.


1. Left identity:
-----------------

Claim: ALL x :: a. ALL f :: a -> Tree b. return x >>= f = f x

Proof: Let 'x :: a' and 'f :: a -> Tree b' be arbitrary. Then, we have that
  
          return x >>= f
        = Leaf x >>= f             -- return.1
        = f x                      -- (>>=).1

       which concludes the proof.

Qed.


2. Right identity:
------------------

Claim: ALL t :: Tree a. t >>= return = t

Proof: 
  Let P(t) := t >>= return = t. 
  We prove ALL t :: Tree a. P(t) by induction.

  Case Leaf: Let x::a be arbitrary.
    We must show that P(Leaf x) holds. Note that

        Leaf x >>= return
      = return x             -- (>>=).1
      = x                    -- return.1

    which concludes this case.

  Case Node: Let l::Tree a and r::Tree a be arbitrary.
    We must show that P(Node l r) holds provided that P(l) and P(r) hold.
    Note that

        Node l r >>= return
      = Node (l >>= return) (r >>= return)    -- (>>=).2
      = Node l r                              -- P(l) and P(r)

    which concludes this case.

Qed.


3. Associativity:
-----------------

Claim: ALL t :: Tree a. ALL f :: a -> Tree b. ALL g :: b -> Tree c.
         (t >>= f) >>= g = t >>= (\x -> f x >>= g)

Proof: 
  Let P(t) := ALL f :: a -> Tree b. ALL g :: b -> Tree c.
                 (t >>= f) >>= g = t >>= (\x -> f x >>= g).
  We prove ALL t :: Tree a. P(t) by induction over t.

  Case Leaf: Let x::a be arbitrary. We must show that P(Leaf x) holds. Note that

        (Leaf x >>= f) >>= g
      = f x >>= g                        -- (>>=).1
      = Leaf x >>= (\x -> f x >>= g)     -- (>>=).1 backwards

    which concludes this case.

  Case Node: Let l::Tree a and r::Tree a be arbitrary. We must show that
    P(Node l r) holds provided that P(l) and P(r) hold. Note that

        (Node l r >>= f) >>= g
      = (Node (l >>= f) (r >>= f)) >>= g                           -- (>>=).2
      = Node ((l >>= f) >>= g) ((r >>= f) >>= g)                   -- (>>=).2
      = Node (l >>= (\x -> f x >>= g)) (r >>= (\x -> f x >>= g))   -- P(l) and P(r)
      = Node l r >>= (\x -> f x >>= g)                             -- (>>=).2 backwards
    
    which concludes this case.

Qed.


-- Assignment 2
---------------

a.i) lazy

          (\x. \y. x y) (\z. y z)
      ~~> (\y'. (\z. y z) y')

a.ii) eager

          (\x. \y. x y) (\z. y z)
      ~~> (\y'. (\z. y z) y')
      ~~> (\y'. y y')


b.i) lazy

          ((\f. \x. x f) (\r. r)) ((\y. \z. y z) (\a. a))
      ~~> (\x. x (\r. r)) ((\y. \z. y z) (\a. a))
      ~~> ((\y. \z. y z) (\a. a)) (\r. r)
      ~~> (\z. (\a. a) z) (\r. r)
      ~~> (\a. a) (\r. r)
      ~~> (\r. r)

b.ii) eager

          ((\f. \x. x f) (\r. r)) ((\y. \z. y z) (\a. a))
      ~~> (\x. x (\r. r)) ((\y. \z. y z) (\a. a))
      ~~> (\x. x (\r. r)) (\z. (\a. a) z)
      ~~> (\x. x (\r. r)) (\z. z)
      ~~> (\z. z) (\r. r)
      ~~> (\r. r)



-- Assignment 3
---------------

a)

  1. (\x y z -> (y x, x z)) :: (c -> b) -> ((c -> b) -> a) -> c -> (a, b)
  2. (\x y -> x (\z -> y)) :: ((c -> a) -> b) -> a -> b
  3. map (1 :)  :: Num a => [[a]] -> [[a]]
  4. (\x ys -> inits (map (x <) ys)) :: Ord a => a -> [a] -> [[Bool]]

b)

Let G := y : a, x : a -> Int 

                                           ------------------ Ax  ----------- Ax
                                           G |- x :: a -> Int     G |- y :: a
                             ------ Int    ---------------------------------- App
                             G |- 1        G |- x y :: Int
                             ------------------------------------------------ Op +
                                 G |- 1 + x y :: Int
--------------------- True       -------------------------------------------- Abs
y : a |- True :: Bool            y : a |- \x -> 1 + x y :: (a -> Int) -> Int
----------------------------------------------------------------------------- Tup
y : a |- (True, \x -> 1 + x y) :: (Bool, (a -> Int) -> Int)
----------------------------------------------------------------------------- Abs
|- (\y -> (True, \x -> 1 + x y)) :: a -> (Bool, (a -> Int) -> Int)



-- Assignment 4
---------------


Recall the definitions:

  addShifted :: [Int] -> Int -> [Int]
  addShifted []     i = [i]                       -- addShifted.1
  addShifted (x:xs) i = (x + i):addShifted xs x   -- addShifted.2

  sum :: [Int] -> Int
  sum []     = 0                                  -- sum.1
  sum (x:xs) = x + sum xs                         -- sum.2

Claim: ALL xs::[Int]. ALL i::Int. sum (addShifted xs i) = i + 2 * sum xs.

Proof: 
  Let P(xs) := ALL i::Int. sum (addShifted xs i) = i + 2 * sum xs.
  We show ALL xs::[Int]. P(xs) by induction.

Base Case: 
  Show P([]).
  Let i::Int be arbitrary.

    sum (addShifted [] i)
  = sum [i]              -- addShifted.1
  = i + sum []           -- sum.2
  = i + 0                -- sum.1
  = i + 2 * 0            -- arith
  = i + 2 * sum []       -- sum.1

Step Case: 
  Let x :: a and xs :: [a] be arbitrary. Assume P(xs).
  Show P(x:xs). Let i::Int be arbitrary. 

    sum (addShifted (x:xs) i)
  = sum ((x+i):addShifted xs x)   -- addShifted.1
  = (x+i) + sum (addShifted xs x) -- sum.2
  = (x+i) + x + 2*sum xs          -- P(xs) where i is instantiated with x
  = i + 2*(x+sum xs)              -- arith
  = i + 2*(sum (x:xs))            -- sum.2

Qed.


-- Assignment 5
---------------

a)

We use a local function go with an additional argument that keeps track of the
open parentheses in a stack. We use the simple construction which hardcodes
the corresponding pairs of opening and closing parentheses.

> match :: String -> Bool
> match s = go s []
>   where go []       []           = True
>         go []       (_:_)        = False
>         go ('(':cs) parens       = go cs ('(':parens)
>         go ('{':cs) parens       = go cs ('{':parens)
>         go (')':cs) ('(':parens) = go cs parens
>         go (')':cs) (_:parens)   = False
>         go ('}':cs) ('{':parens) = go cs parens
>         go ('}':cs) (_:parens)   = False
>         go (_:cs)   parens       = go cs parens

The two test cases from the exercise sheet.

> t1_1 = match "(a{b}cd)()ef"
> t1_2 = match "(xy{z}}"

b)

The first version uses a function that accumulates a rising list in its
first argument.

> risers :: Ord a => [a] -> [[a]]
> risers []    = []
> risers (x:xs)= collect [x] xs
>   where 
>   collect rising []    = reverse rising : []
>   collect rising (y:ys)
>     | head rising <= y =                  collect (y:rising) ys
>     | otherwise        = reverse rising : collect [y]        ys

The second version uses span and zip

> risers2 :: Ord a => [a] -> [[a]]
> risers2 []       = []
> risers2 xs@(x:_) = (x : map snd rise) : risers2 (map snd rest)
>   where (rise,rest) = span (uncurry (<=)) (zip xs (tail xs))

There is also a version using foldr.

> risers3 :: Ord a => [a] -> [[a]]
> risers3 = foldr go []
>  where go x []          = [[x]]
>        go x ((y:ys):zs)
>         | x <= y        = (x:y:ys):zs
>         | otherwise     = [x]:(y:ys):zs



-- Assignment 6
---------------

a)

> data Expr = Varx           -- ^ The one variable 'x' that we allow in expressions.
>           | ConstE         -- ^ The constant 'e'.
>           | Num Double     -- ^ Some real number.
>           | Mul Expr Expr 
>           | Add Expr Expr 
>           | Div Expr Expr 
>           | Pow Expr Expr 
>           | Exp Expr 
>           | Log Expr

This instance is not part of the answer, but it makes the expressions much
easier to read in experiments.

> instance Show Expr where
>     show (Mul e1 e2)= "(" ++ (show e1) ++ "*" ++ (show e2) ++ ")"
>     show (Add e1 e2)= "(" ++ (show e1) ++ "+" ++ (show e2) ++ ")"
>     show (Div e1 e2)= "(" ++ (show e1) ++ "/" ++ (show e2) ++ ")"
>     show (Pow e1 e2) = "(" ++ (show e1) ++ ")^" ++ (show e2)
>     show (Exp e1) = "e^(" ++ (show e1) ++ ")"
>     show (Log e1) = "ln(" ++ (show e1) ++ ")"
>     show Varx = "x"
>     show ConstE = "e"
>     show (Num x) = show x


b)

> diff :: Expr -> Expr
> diff Varx        = Num 1.0
> diff ConstE      = Num 0.0
> diff (Num _)     = Num 0.0
> diff (Add e1 e2) = Add (diff e1) (diff e2)
> diff (Mul e1 e2) = Add (Mul (diff e1) e2) (Mul e1 (diff e2))
> diff (Log e)     = Div (diff e) e
> diff (Exp e)     = Mul (diff e) (Exp e)
> diff (Div e1 e2) = diff (Mul e1 (Pow e2 (Num (-1.0))))
> diff (Pow e1 e2) = diff (Exp (Mul e2 (Log e1)))

a simple test expression

> test_e1 = Exp (Mul (Num 10) Varx)

